@@ -1,5 +1,5 @@
 //
-// $Id: sphinxexcerpt.cpp 1276 2008-05-19 20:31:26Z shodan $
+// $Id: sphinxexcerpt.cpp 1729 2009-03-05 12:55:07Z xale $
 //
 
 //
@@ -79,6 +79,20 @@ public:
 		}
 	};
 
+	enum KeywordStar_e
+	{
+		STAR_NONE	= 0,
+		STAR_FRONT	= 1 << 0,
+		STAR_BACK	= 1 << 1,
+		STAR_BOTH	= STAR_FRONT | STAR_BACK
+	};
+
+	struct Keyword_t
+	{
+		int		m_uStar;
+		int		m_iWord;
+	};
+
 protected:
 	CSphVector<Token_t>		m_dTokens;		///< source text tokens
 	CSphVector<Token_t>		m_dWords;		///< query words tokens
@@ -93,12 +107,11 @@ protected:
 	bool					m_bExactPhrase;
 
 protected:
-	bool					TokensMatch ( const Token_t & a, const Token_t & b);
-
 	void					CalcPassageWeight ( const CSphVector<int> & dPassage, Passage_t & tPass, int iMaxWords, int iWordCountCoeff );
 	bool					ExtractPassages ( const ExcerptQuery_t & q );
 	bool					ExtractPhrases ( const ExcerptQuery_t & q );
 
+	void					HighlightPhrase ( const ExcerptQuery_t & q, int iStart, int iEnd );
 	void					HighlightAll ( const ExcerptQuery_t & q );
 	void					HighlightStart ( const ExcerptQuery_t & q );
 	bool					HighlightBestPassages ( const ExcerptQuery_t & q );
@@ -106,7 +119,8 @@ protected:
 	void					ResultEmit ( const char * sLine );
 	void					ResultEmit ( const Token_t & sTok );
 
-	void					AddJunk ( int iStart, int iLength );
+	void					AddJunk ( int iStart, int iLength, int iBoundary );
+	void					AddBoundary ();
 };
 
 /////////////////////////////////////////////////////////////////////////////
@@ -126,17 +140,27 @@ inline bool operator < ( const ExcerptGen_c::Passage_t & a, const ExcerptGen_c::
 	return a.GetWeight() > b.GetWeight();
 }
 
-
 ExcerptGen_c::ExcerptGen_c ()
 {
 	m_bExactPhrase = false;
 }
 
+void ExcerptGen_c::AddBoundary()
+{
+	m_dTokens.Resize ( m_dTokens.GetLength () + 1 );
+	Token_t & tLast = m_dTokens.Last ();
+	tLast.m_eType = TOK_BREAK;
+	tLast.m_iStart = 0;
+	tLast.m_iLengthBytes = 0;
+	tLast.m_iWordID = 0;
+	tLast.m_uWords = 0;
+}
 
-void ExcerptGen_c::AddJunk ( int iStart, int iLength )
+void ExcerptGen_c::AddJunk ( int iStart, int iLength, int iBoundary )
 {
 	int iChunkStart = iStart;
-	
+	int iSaved = 0;
+
 	for ( int i = iStart; i < iStart+iLength; i++ )
 		if ( sphIsSpace ( m_sBuffer.cstr () [i] ) != sphIsSpace ( m_sBuffer.cstr () [iChunkStart] ) )
 		{
@@ -146,8 +170,16 @@ void ExcerptGen_c::AddJunk ( int iStart, int iLength )
 			tLast.m_iStart	= iChunkStart;
 			tLast.m_iLengthBytes = i - iChunkStart;
 			tLast.m_iWordID = 0;
+			tLast.m_uWords = 0;
 
 			iChunkStart = i;
+			iSaved += tLast.m_iLengthBytes;
+
+			if ( iBoundary != -1 && iSaved > iBoundary - iStart )
+			{
+				AddBoundary();
+				iBoundary = -1;
+			}
 		}
 
 	m_dTokens.Resize ( m_dTokens.GetLength () + 1 );
@@ -156,35 +188,78 @@ void ExcerptGen_c::AddJunk ( int iStart, int iLength )
 	tLast.m_iStart	= iChunkStart;
 	tLast.m_iLengthBytes = iStart + iLength - iChunkStart;
 	tLast.m_iWordID = 0;
+	tLast.m_uWords = 0;
+
+	if ( iBoundary != -1 ) AddBoundary();
 }
 
 
 char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict, ISphTokenizer * pTokenizer )
 {
 	m_dTokens.Reserve ( 1024 );
-
 	m_sBuffer = q.m_sSource;
 
-	pTokenizer->SetBuffer ( (BYTE*)q.m_sSource.cstr (), strlen ( q.m_sSource.cstr () ) );
+	const bool bUtf8 = pTokenizer->IsUtf8();
+
+	// tokenize query words
+	int iWordsLength = strlen ( q.m_sWords.cstr() );
+
+	CSphVector<char> dKwBuffer ( iWordsLength );
+	CSphVector<Keyword_t> dKeywords;
+	dKeywords.Reserve ( MAX_HIGHLIGHT_WORDS );
+
 	BYTE * sWord;
+	int iKwIndex = 0;
+
+	pTokenizer->SetBuffer ( (BYTE*)q.m_sWords.cstr(), iWordsLength );
+	while ( ( sWord = pTokenizer->GetToken() ) != NULL )
+	{
+		SphWordID_t iWord = pDict->GetWordID ( sWord );
+		if ( iWord )
+		{
+			m_dWords.Resize ( m_dWords.GetLength () + 1 );
+			Token_t & tLast = m_dWords.Last ();
+			tLast.m_eType = TOK_WORD;
+			tLast.m_iWordID = iWord;
+			tLast.m_iLengthBytes = strlen ( (const char *)sWord );
+			tLast.m_iLengthCP = bUtf8 ? sphUTF8Len ( (const char *)sWord ) : tLast.m_iLengthBytes;
+
+			// store keyword
+			dKeywords.Resize( dKeywords.GetLength() + 1 );
+			Keyword_t & kwLast = dKeywords.Last ();
+
+			// find stars
+			bool bStarBack = *pTokenizer->GetTokenEnd() == '*';
+			bool bStarFront = ( pTokenizer->GetTokenStart() != pTokenizer->GetBufferPtr() ) &&
+				pTokenizer->GetTokenStart()[-1] == '*';
+			kwLast.m_uStar = ( bStarFront ? STAR_FRONT : 0 ) | ( bStarBack ? STAR_BACK : 0 );
+
+			// store token
+			const int iEndIndex = iKwIndex + tLast.m_iLengthBytes + 1;
+			dKwBuffer.Resize ( iEndIndex );
+			kwLast.m_iWord = iKwIndex;
+			strcpy ( &dKwBuffer [ iKwIndex ], (const char *)sWord );
+			iKwIndex = iEndIndex;
+
+			if ( m_dWords.GetLength() == MAX_HIGHLIGHT_WORDS )
+				break;
+		}
+	}
+
+	// tokenize document
+	pTokenizer->SetBuffer ( (BYTE*)q.m_sSource.cstr (), strlen ( q.m_sSource.cstr () ) );
+
 	const char * pStartPtr = pTokenizer->GetBufferPtr ();
 	const char * pLastTokenEnd = pStartPtr;
+
 	while ( ( sWord = pTokenizer->GetToken() ) != NULL )
 	{
 		const char * pTokenStart = pTokenizer->GetTokenStart ();
 
 		if ( pTokenStart != pStartPtr )
-			AddJunk ( pLastTokenEnd - pStartPtr, pTokenStart - pLastTokenEnd );
-
-		if ( pTokenizer->GetBoundary() )
-		{
-			m_dTokens.Resize ( m_dTokens.GetLength () + 1 );
-			Token_t & tLast = m_dTokens.Last ();
-			tLast.m_eType   = TOK_BREAK;
-			tLast.m_iStart  = 0;
-			tLast.m_iLengthBytes = 0;
-			tLast.m_iWordID = 0;
-		}
+			AddJunk ( pLastTokenEnd - pStartPtr,
+					  pTokenStart - pLastTokenEnd,
+					  pTokenizer->GetBoundary() ? pTokenizer->GetBoundaryOffset() : -1 );
 
 		SphWordID_t iWord = pDict->GetWordID ( sWord );
 
@@ -196,19 +271,63 @@ char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict,
 		tLast.m_iStart  = pTokenStart - pStartPtr;
 		tLast.m_iLengthBytes = pLastTokenEnd - pTokenStart;
 		tLast.m_iWordID = iWord;
+		tLast.m_uWords = 0;
+
+		// fill word mask
+		if ( iWord )
+		{
+			bool bMatch = false;
+			int iOffset;
+
+			ARRAY_FOREACH ( nWord, m_dWords )
+			{
+				const char * keyword = &dKwBuffer [ dKeywords[nWord].m_iWord ];
+				const Token_t & token = m_dWords[nWord];
+
+				switch ( dKeywords[nWord].m_uStar )
+				{
+				case STAR_NONE:
+					bMatch = iWord == token.m_iWordID;
+					break;
+
+				case STAR_FRONT:
+					iOffset = tLast.m_iLengthBytes - token.m_iLengthBytes;
+					bMatch = (iOffset >= 0) &&
+						( memcmp( keyword, sWord + iOffset, token.m_iLengthBytes ) == 0 );
+					break;
+
+				case STAR_BACK:
+					bMatch = ( tLast.m_iLengthBytes >= token.m_iLengthBytes ) &&
+						( memcmp( keyword, sWord, token.m_iLengthBytes ) == 0 );
+					break;
+
+				case STAR_BOTH:
+					bMatch = strstr( (const char *)sWord, keyword ) != NULL;
+					break;
+				}
+
+				if ( bMatch )
+					tLast.m_uWords |= (1UL << nWord);
+			}
+		}
 	}
 
 	// last space if any
 	if ( pLastTokenEnd != pTokenizer->GetBufferEnd () )
-		AddJunk ( pLastTokenEnd - pStartPtr, pTokenizer->GetBufferEnd () - pLastTokenEnd );
-
+	{
+		int iOffset = pTokenizer->GetBoundary() ? pTokenizer->GetBoundaryOffset() : -1;
+		AddJunk ( pLastTokenEnd - pStartPtr, pTokenizer->GetBufferEnd () - pLastTokenEnd, iOffset );
+	}
+	
 	m_dTokens.Resize ( m_dTokens.GetLength () + 1 );
 	Token_t & tLast = m_dTokens.Last ();
 	tLast.m_eType   = TOK_NONE;
 	tLast.m_iStart  = 0;
 	tLast.m_iLengthBytes = 0;
 	tLast.m_iWordID = 0;
-	
+	tLast.m_uWords = 0;
+
+	// sum token lengths
 	int iSourceCodes = 0;
 	ARRAY_FOREACH ( i, m_dTokens )
 	{
@@ -216,34 +335,21 @@ char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict,
 
 		if ( m_dTokens [i].m_iLengthBytes )
 		{
-			m_dTokens [i].m_iLengthCP = sphUTF8Len ( m_sBuffer.SubString ( m_dTokens [i].m_iStart, m_dTokens [i].m_iLengthBytes ).cstr () );
-			iSourceCodes += m_dTokens [i].m_iLengthCP;
+			if ( bUtf8 )
+			{
+				int iLen = sphUTF8Len ( m_sBuffer.SubString ( m_dTokens[i].m_iStart, m_dTokens[i].m_iLengthBytes ).cstr() );
+				m_dTokens[i].m_iLengthCP = iLen;
+			}
+			else
+				m_dTokens[i].m_iLengthCP = m_dTokens[i].m_iLengthBytes;
+			iSourceCodes += m_dTokens[i].m_iLengthCP;
 		}
 		else
 			m_dTokens [i].m_iLengthCP = 0;
 	}
 
-	pTokenizer->SetBuffer ( (BYTE*)q.m_sWords.cstr (), strlen ( q.m_sWords.cstr () ) );
-
-	while ( ( sWord = pTokenizer->GetToken() ) != NULL )
-	{
-		SphWordID_t iWord = pDict->GetWordID ( sWord );
-		if ( iWord )
-		{
-			m_dWords.Resize ( m_dWords.GetLength () + 1 );
-			Token_t & tLast = m_dWords.Last ();
-			tLast.m_eType = TOK_WORD;
-			tLast.m_iWordID = iWord;
-			tLast.m_iLengthCP = sphUTF8Len ( (const char*)sWord );
-		}
-	}
-
 	m_bExactPhrase = q.m_bExactPhrase && ( m_dWords.GetLength()>1 );
 
-	// truncate the array
-	if ( m_dWords.GetLength()>MAX_HIGHLIGHT_WORDS )
-		m_dWords.Resize ( MAX_HIGHLIGHT_WORDS );
-
 	// assign word weights
 	ARRAY_FOREACH ( i, m_dWords )
 		m_dWords[i].m_iWeight = m_dWords[i].m_iLengthCP; // FIXME! should obtain freqs from dict
@@ -253,15 +359,6 @@ char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict,
 	m_dResult.Resize ( 0 );
 	m_iResultLen = 0;
 
-	// calc matching word masks
-	ARRAY_FOREACH ( iTok, m_dTokens )
-	{
-		m_dTokens[iTok].m_uWords = 0;
-		ARRAY_FOREACH ( iWord, m_dWords )
-			if ( TokensMatch ( m_dTokens[iTok], m_dWords[iWord] ) )
-				m_dTokens[iTok].m_uWords |= (1UL<<iWord);
-	}
-
 	// do highlighting
 	if ( q.m_iLimit<=0 || q.m_iLimit>iSourceCodes )
 	{
@@ -273,10 +370,6 @@ char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict,
 			HighlightStart ( q );
 	}
 
-	// cleanup
-	m_dTokens.Reset ();
-	m_dWords.Reset ();
-
 	// alloc, fill and return the result
 	m_dResult.Add ( 0 );
 	char * pRes = new char [ m_dResult.GetLength() ];
@@ -287,63 +380,53 @@ char * ExcerptGen_c::BuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict,
 }
 
 
+void ExcerptGen_c::HighlightPhrase ( const ExcerptQuery_t & q, int iTok, int iEnd )
+{
+	while ( iTok<=iEnd )
+	{
+		while ( iTok<=iEnd && !m_dTokens[iTok].m_uWords )
+			ResultEmit ( m_dTokens[iTok++] );
+
+		if ( iTok>iEnd )
+			break;
+
+		bool bMatch = true;
+		int iWord = 0;
+		int iStart = iTok;
+		while ( iWord<m_dWords.GetLength() )
+		{
+			if ( ( iTok > iEnd ) ||
+				 !( m_dTokens[iTok].m_eType==TOK_SPACE || m_dTokens[iTok].m_uWords == ( 1UL<<iWord++ ) ) )
+			{
+				bMatch = false;
+				break;
+			}
+			iTok++;
+		}
+
+		if ( !bMatch )
+		{
+			ResultEmit ( m_dTokens[iStart] );
+			iTok = iStart + 1;
+			continue;
+		}
+
+		ResultEmit ( q.m_sBeforeMatch.cstr() );
+		while ( iStart<iTok )
+			ResultEmit ( m_dTokens [ iStart++ ] );
+		ResultEmit ( q.m_sAfterMatch.cstr() );
+	}
+}
+
+
 void ExcerptGen_c::HighlightAll ( const ExcerptQuery_t & q )
 {
 	bool bOpen = false;
 	const int iMaxTok = m_dTokens.GetLength()-1; // skip last one, it's TOK_NONE
 
 	if ( m_bExactPhrase )
-	{
-		// exact phrase
-		for ( int iCur=0; iCur<iMaxTok; )
-		{
-			// skip non-opening words
-			while ( iCur<iMaxTok && !( m_dTokens[iCur].m_uWords & 1 ))
-				ResultEmit ( m_dTokens[iCur++] );
-
-			// check if we have enough words left
-			if ( iCur+m_dWords.GetLength()-1>=iMaxTok )
-			{
-				// not enough, just copy the tail
-				while ( iCur<iMaxTok )
-					ResultEmit ( m_dTokens[iCur++] );
-				break;
-			}
-
-			// lookahead
-			assert ( iCur>=0 && iCur<iMaxTok );
-			assert ( m_dTokens[iCur].m_uWords & 1 );
-
-			int iLookahead = 1; // current lookahead position
-			int iMatched = 1; // phrase words matched so far
-			while ( iCur+iLookahead<iMaxTok && iMatched<m_dWords.GetLength() )
-			{
-				const Token_t & tTok = m_dTokens[iCur+iLookahead];
-				if ( !tTok.m_uWords )
-				{
-					iLookahead++;
-					continue;
-				}
-
-				if (!( tTok.m_uWords & (1<<iMatched) ))
-					break;
-
-				iLookahead++;
-				iMatched++;
-			}
-
-			// emit looked-ahead tokens
-			if ( iMatched==m_dWords.GetLength() )
-				ResultEmit ( q.m_sBeforeMatch.cstr() );
-
-			while ( iLookahead-- )
-				ResultEmit ( m_dTokens[iCur++] );
-
-			if ( iMatched==m_dWords.GetLength() )
-				ResultEmit ( q.m_sAfterMatch.cstr() );
-		}
-
-	} else
+		HighlightPhrase ( q, 0, iMaxTok-1 );
+	else
 	{
 		// bag of words
 		for ( int iTok=0; iTok<iMaxTok; iTok++ )
@@ -374,13 +457,6 @@ void ExcerptGen_c::HighlightStart ( const ExcerptQuery_t & q )
 	ResultEmit ( q.m_sChunkSeparator.cstr() );
 }
 
-
-bool ExcerptGen_c::TokensMatch ( const Token_t & a, const Token_t & b )
-{
-	return a.m_iWordID==b.m_iWordID;
-}
-
-
 void ExcerptGen_c::ResultEmit ( const char * sLine )
 {
 	while ( *sLine )
@@ -474,102 +550,81 @@ bool ExcerptGen_c::ExtractPassages ( const ExcerptQuery_t & q )
 	int iMaxWords = 2*q.m_iAround+1;
 	int iLCSThresh = m_bExactPhrase ? m_dWords.GetLength()*iMaxWords : 0;
 
-	ARRAY_FOREACH ( iTok, m_dTokens )
+	// setup initial window
+	ARRAY_FOREACH ( i, m_dTokens )
 	{
+		const Token_t & tToken = m_dTokens[i];
+
 		// skip starting whitespace
-		if ( tPass.m_iTokens==0 && m_dTokens[iTok].m_eType!=TOK_WORD )
+		if ( tPass.m_iTokens == 0 && tToken.m_eType != TOK_WORD )
 		{
 			tPass.m_iStart++;
 			continue;
 		}
 
+		// stop when the window is large enough
+		if ( tPass.m_iCodes + tToken.m_iLengthCP > q.m_iLimit || dPass.GetLength() == iMaxWords )
+			break;
+
 		// got token, update passage
 		tPass.m_iTokens++;
-		tPass.m_iCodes += m_dTokens[iTok].m_iLengthCP;
+		tPass.m_iCodes += tToken.m_iLengthCP;
 
-		if ( m_dTokens[iTok].m_eType!=TOK_WORD )
-			continue;
-
-		// got word
-		dPass.Add ( iTok );
-		tPass.m_uWords |= m_dTokens[iTok].m_uWords;
-
-		// check if we're done
-		if ( dPass.GetLength()>=iMaxWords )
-			break;
+		if ( tToken.m_eType == TOK_WORD )
+		{
+			dPass.Add(i);
+			tPass.m_uWords |= tToken.m_uWords;
+		}
 	}
 
-	if ( tPass.m_uWords )
-	{
-		CalcPassageWeight ( dPass, tPass, iMaxWords, 0 );
-		if ( tPass.m_iMaxLCS>=iLCSThresh )
-			m_dPassages.Add ( tPass );
-	}
-
-	// my lovely update-and-submit loop
+	// move our window until the end of document
+	const int iCount = m_dTokens.GetLength();
 	for ( ;; )
 	{
-		int iAdd = tPass.m_iStart + tPass.m_iTokens;
-		if ( iAdd>=m_dTokens.GetLength() )
-			break;
-
-		// remove opening word
-		assert ( m_dTokens[tPass.m_iStart].m_eType==TOK_WORD );
-		tPass.m_iTokens--;
-		tPass.m_iCodes -= m_dTokens [ tPass.m_iStart ].m_iLengthCP;
-		tPass.m_iStart++;
-		dPass.Remove ( 0 ); // FIXME! OPTIMIZE! make this a cyclic buffer
-
-		// remove opening non-words
-		while ( m_dTokens[tPass.m_iStart].m_eType!=TOK_WORD )
-		{
-			tPass.m_iTokens--;
-			tPass.m_iCodes -= m_dTokens [ tPass.m_iStart ].m_iLengthCP;
-			tPass.m_iStart++;
-		}
-
-		// add trailing non-words
-		assert ( tPass.m_iStart+tPass.m_iTokens==iAdd );
-		while ( iAdd<m_dTokens.GetLength() && m_dTokens[iAdd].m_eType!=TOK_WORD )
-		{
-			tPass.m_iTokens++;
-			tPass.m_iCodes += m_dTokens[iAdd].m_iLengthCP;
-			iAdd++;
-		}
-		if ( iAdd>=m_dTokens.GetLength() )
-			break;
-
-		// add trailing word
-		assert ( tPass.m_iStart+tPass.m_iTokens==iAdd );
-		assert ( m_dTokens[iAdd].m_eType==TOK_WORD );
-		tPass.m_iTokens++;
-		tPass.m_iCodes += m_dTokens[iAdd].m_iLengthCP;
-		dPass.Add ( iAdd );
-
 		// re-weight current passage, and check if it matches
 		CalcPassageWeight ( dPass, tPass, iMaxWords, 0 );
-		if ( !tPass.m_uWords || tPass.m_iMaxLCS<iLCSThresh )
-			continue;
-
-		// if it's the very first one, do add
-		if ( !m_dPassages.GetLength() )
+		if ( tPass.m_uWords && tPass.m_iMaxLCS >= iLCSThresh )
 		{
-			m_dPassages.Add ( tPass );
-			continue;
+			// if it's the very first one, do add
+			if ( !m_dPassages.GetLength() )
+				m_dPassages.Add ( tPass );
+			else
+			{
+				// check if it's new or better
+				Passage_t & tLast = m_dPassages.Last();
+				if ( tLast.m_uWords!=tPass.m_uWords || tLast.m_iStart + tLast.m_iTokens - 1 < tPass.m_iStart )
+					m_dPassages.Add ( tPass ); // new
+				else if ( tLast.GetWeight() < tPass.GetWeight() )
+					tLast = tPass;  // better
+			}
 		}
 
-		// check if it's new or better
-		Passage_t & tLast = m_dPassages.Last();
-		if ( tLast.m_uWords!=tPass.m_uWords
-			|| tLast.m_iStart+tLast.m_iTokens-1 < tPass.m_iStart )
+		int iToken = tPass.m_iStart + tPass.m_iTokens;
+		assert ( iToken <= iCount );
+		if ( iToken == iCount ) break;
+
+		// add another word
+		for ( ; iToken < iCount; iToken++ )
 		{
-			// new
-			m_dPassages.Add ( tPass );
-		} else
+			tPass.m_iTokens++;
+			tPass.m_iCodes += m_dTokens[iToken].m_iLengthCP;
+			if ( m_dTokens[iToken].m_eType == TOK_WORD )
+			{
+				dPass.Add ( iToken );
+				break;
+			}
+		}
+		if ( iToken == iCount ) continue;
+
+		// drop front tokens until the window fits into both word and CP limits
+		while ( ( tPass.m_iCodes > q.m_iLimit || dPass.GetLength() > iMaxWords ) && tPass.m_iTokens != 1 )
 		{
-			// better
-			if ( tLast.GetWeight()<tPass.GetWeight() )
-				tLast = tPass;
+			if ( m_dTokens[tPass.m_iStart].m_eType == TOK_WORD )
+				dPass.Remove ( 0 );
+
+			tPass.m_iCodes -= m_dTokens[tPass.m_iStart].m_iLengthCP;
+			tPass.m_iTokens--;
+			tPass.m_iStart++;
 		}
 	}
 
@@ -579,6 +634,9 @@ bool ExcerptGen_c::ExtractPassages ( const ExcerptQuery_t & q )
 
 bool ExcerptGen_c::ExtractPhrases ( const ExcerptQuery_t & )
 {
+	int iMaxWords = 100;
+	int iLCSThresh = m_bExactPhrase ? m_dWords.GetLength()*iMaxWords : 0;
+
 	int iStart = 0;
 	DWORD uWords = 0;
 
@@ -587,10 +645,7 @@ bool ExcerptGen_c::ExtractPhrases ( const ExcerptQuery_t & )
 		// phrase boundary found, go flush
 		if ( m_dTokens[iTok].m_eType==TOK_BREAK || m_dTokens[iTok].m_eType==TOK_NONE )
 		{
-			// where's my ending token
-			int iEnd = iTok;
-			if ( m_dTokens[iTok].m_eType==TOK_NONE )
-				iEnd--;
+			int iEnd = iTok - 1;
 
 			// emit non-empty phrases with matching words as passages
 			if ( iStart<iEnd && uWords!=0 )
@@ -609,18 +664,16 @@ bool ExcerptGen_c::ExtractPhrases ( const ExcerptQuery_t & )
 						dPass.Add ( i );
 				}
 
-				CalcPassageWeight ( dPass, tPass, 100, 10000 );
-				m_dPassages.Add ( tPass );
+				CalcPassageWeight ( dPass, tPass, iMaxWords, 10000 );
+				if ( tPass.m_iMaxLCS >= iLCSThresh )
+					m_dPassages.Add ( tPass );
+
 			}
 
-			// skip until first word or EOF
-			while ( m_dTokens[iTok].m_eType!=TOK_WORD && m_dTokens[iTok].m_eType!=TOK_NONE )
-				iTok++;
 			if ( m_dTokens[iTok].m_eType==TOK_NONE )
 				break;
 
-			assert ( m_dTokens[iTok].m_eType==TOK_WORD );
-			iStart = iTok;
+			iStart = iTok + 1;
 			uWords = 0;
 		}
 
@@ -651,39 +704,88 @@ bool ExcerptGen_c::HighlightBestPassages ( const ExcerptQuery_t & q )
 	CSphVector<Passage_t> dShow;
 	int iLeft = q.m_iLimit;
 
-	while ( ( q.m_bUseBoundaries || iLeft>0 ) && m_dPassages.GetLength() )
+	if ( ( q.m_bUseBoundaries || iLeft>0 ) && m_dPassages.GetLength() )
 	{
-		// FIXME! use heap instead of sorting again every time?
-		m_dPassages.Sort ();
-		Passage_t & tPass = m_dPassages[0];
-
-		if ( tPass.m_iCodes<=iLeft || q.m_bUseBoundaries )
+		// initial heapify
+		for ( int i=1; i<m_dPassages.GetLength(); i++ )
 		{
-			// add it to the show
-			dShow.Add ( tPass );
-			iLeft -= tPass.m_iCodes;
+			// everything upto i-th is heapified; sift up i-th element
+			for ( int j=i; j!=0 && ( m_dPassages[j] < m_dPassages[j>>1] ); j=j>>1 )
+				Swap ( m_dPassages[j>>1], m_dPassages[j] );
+		}
 
-			// sometimes be need only one best one
-			if ( q.m_bSinglePassage )
-				break;
+		// best passage extraction loop
+		DWORD uNotShown = 1UL << ( m_dWords.GetLength()-1 );
+		while ( m_dPassages.GetLength() )
+		{
+			// this is our hero
+			Passage_t & tPass = m_dPassages[0];
+
+			// emit this passage, if we can
+			DWORD uShownWords = 0;
+			if ( tPass.m_iCodes<=iLeft || q.m_bUseBoundaries )
+			{
+				// add it to the show
+				dShow.Add ( tPass );
+				iLeft -= tPass.m_iCodes;
+				uShownWords = tPass.m_uWords;
+
+				// sometimes we need only one best one
+				if ( q.m_bSinglePassage )
+					break;
+			}
+
+			// promote tail, retire head
+			m_dPassages.RemoveFast ( 0 );
+
+			// sift down former tail
+			int iEntry = 0;
+			for ( ;; )
+			{
+				// select child
+				int iChild = (iEntry<<1) + 1;
+				if ( iChild>=m_dPassages.GetLength() )
+					break;
+
+				// select smallest child
+				if ( iChild+1<m_dPassages.GetLength() )
+					if ( m_dPassages[iChild+1] < m_dPassages[iChild] )
+						iChild++;
+
+				// if smallest child is less than entry, exchange and continue
+				if (!( m_dPassages[iChild]<m_dPassages[iEntry] ))
+					break;
+				Swap ( m_dPassages[iChild], m_dPassages[iEntry] );
+				iEntry = iChild;
+			}
 
 			// we now show some of the query words,
 			// so displaying other passages containing those is less significant,
 			// so let's update all the other weights (and word masks, to avoid updating twice)
-			for ( int i=1; i<m_dPassages.GetLength(); i++ )
-				if ( m_dPassages[i].m_uWords & tPass.m_uWords )
+			// and sift up
+			if ( uNotShown )
+				for ( int i=0; i<m_dPassages.GetLength(); i++ )
 			{
-				DWORD uWords = tPass.m_uWords;
-				for ( int iWord=0; uWords; iWord++, uWords>>=1 )
-					if ( ( uWords & 1 ) && ( m_dPassages[i].m_uWords & ( 1UL<<iWord ) ) )
-						m_dPassages[i].m_iWordsWeight -= m_dWords[iWord].m_iWeight;
+				if ( m_dPassages[i].m_uWords & uShownWords )
+				{
+					// update this passage
+					DWORD uWords = uShownWords;
+					for ( int iWord=0; uWords; iWord++, uWords>>=1 )
+						if ( ( uWords & 1 ) && ( m_dPassages[i].m_uWords & ( 1UL<<iWord ) ) )
+							m_dPassages[i].m_iWordsWeight -= m_dWords[iWord].m_iWeight;
 
-				m_dPassages[i].m_uWords &= ~tPass.m_uWords;
-				assert ( m_dPassages[i].m_iWordsWeight>=0 );
+					m_dPassages[i].m_uWords &= ~uShownWords;
+					assert ( m_dPassages[i].m_iWordsWeight>=0 );
+				}
+
+				// every entry above this is both already updated and properly heapified
+				// we only need to sift up, but we need to sift up *every* entry
+				// because its parent might had been updated this time, breaking heap property
+				for ( int j=i; j!=0 && ( m_dPassages[j] < m_dPassages[j>>1] ); j=j>>1 )
+					Swap ( m_dPassages[j>>1], m_dPassages[j] );
 			}
+			uNotShown &= ~uShownWords;
 		}
-
-		m_dPassages.RemoveFast ( 0 );
 	}
 
 	if ( !dShow.GetLength() )
@@ -764,22 +866,27 @@ bool ExcerptGen_c::HighlightBestPassages ( const ExcerptQuery_t & q )
 		if ( iTok>1+iLast || q.m_bWeightOrder )
 			ResultEmit ( q.m_sChunkSeparator.cstr() );
 
-		while ( iTok<=iEnd )
+		if ( m_bExactPhrase )
 		{
-			if ( iTok>iLast || q.m_bWeightOrder )
+			HighlightPhrase ( q, iTok, iEnd );
+		}
+		else // !m_bExactPhrase
+		{
+			while ( iTok<=iEnd )
 			{
-				// FIXME! glue
-				if ( m_dTokens[iTok].m_uWords )
+				if ( iTok>iLast || q.m_bWeightOrder )
 				{
-					ResultEmit ( q.m_sBeforeMatch.cstr() );
-					ResultEmit ( m_dTokens[iTok] );
-					ResultEmit ( q.m_sAfterMatch.cstr() );
-				} else
-				{
-					ResultEmit ( m_dTokens[iTok] );
+					if ( m_dTokens[iTok].m_uWords )
+					{
+						ResultEmit ( q.m_sBeforeMatch.cstr() );
+						ResultEmit ( m_dTokens[iTok] );
+						ResultEmit ( q.m_sAfterMatch.cstr() );
+					}
+					else
+						ResultEmit ( m_dTokens[iTok] );
 				}
+				iTok++;
 			}
-			iTok++;
 		}
 
 		iLast = iEnd;
@@ -799,5 +906,5 @@ char * sphBuildExcerpt ( const ExcerptQuery_t & q, CSphDict * pDict, ISphTokeniz
 }
 
 //
-// $Id: sphinxexcerpt.cpp 1276 2008-05-19 20:31:26Z shodan $
+// $Id: sphinxexcerpt.cpp 1729 2009-03-05 12:55:07Z xale $
 //
