@@ -1,5 +1,5 @@
 //
-// $Id: tests.cpp 1434 2008-09-09 15:35:07Z shodan $
+// $Id: tests.cpp 1716 2009-03-02 15:42:35Z klirichek $
 //
 
 //
@@ -13,6 +13,7 @@
 
 #include "sphinx.h"
 #include "sphinxexpr.h"
+#include "sphinxutils.h"
 #include "sphinxquery.h"
 #include <math.h>
 
@@ -33,7 +34,7 @@ bool CreateSynonymsFile ( const char * sMagic )
 		"   AT & T => AT & T  \n"
 		"standarten fuehrer => Standartenfuehrer\n"
 		"standarten fuhrer  => Standartenfuehrer\n"
-		"OS/2 => OS/2\n" 
+		"OS/2 => OS/2\n"
 		"Ms-Dos => MS-DOS\n"
 		"MS DOS => MS-DOS\n"
 		"feat. => featuring\n"
@@ -54,18 +55,19 @@ bool CreateSynonymsFile ( const char * sMagic )
 ISphTokenizer * CreateTestTokenizer ( bool bUTF8, bool bSynonyms, bool bEscaped = false )
 {
 	CSphString sError;
-	ISphTokenizer * pTokenizer = bUTF8 ? sphCreateUTF8Tokenizer () : sphCreateSBCSTokenizer ();
+	CSphTokenizerSettings tSettings;
+	tSettings.m_iType = bUTF8 ? TOKENIZER_UTF8 : TOKENIZER_SBCS;
+	tSettings.m_iMinWordLen = 2;
+	ISphTokenizer * pTokenizer = ISphTokenizer::Create ( tSettings, sError );
 	assert ( pTokenizer->SetCaseFolding ( "-, 0..9, A..Z->a..z, _, a..z, U+80..U+FF", sError ) );
-	pTokenizer->SetMinWordLen ( 2 );
 	pTokenizer->AddSpecials ( "!-" );
 	if ( bSynonyms )
 		assert ( pTokenizer->LoadSynonyms ( g_sTmpfile, sError ) );
 
 	if ( bEscaped )
-	{	
+	{
 		ISphTokenizer * pOldTokenizer = pTokenizer;
 		pTokenizer = pTokenizer->Clone ( true );
-		pTokenizer->SetMinWordLen ( 2 );
 		SafeDelete ( pOldTokenizer );
 	}
 
@@ -75,7 +77,7 @@ ISphTokenizer * CreateTestTokenizer ( bool bUTF8, bool bSynonyms, bool bEscaped
 
 void TestTokenizer ( bool bUTF8 )
 {
-	const char * sPrefix = bUTF8 
+	const char * sPrefix = bUTF8
 		? "testing UTF8 tokenizer"
 		: "testing SBCS tokenizer";
 
@@ -131,6 +133,9 @@ void TestTokenizer ( bool bUTF8 )
 			"3", "\\thephone",					"thephone",  NULL,
 			"3", "the\\!phone",					"the", "phone", NULL,
 			"3", "\\!phone",					"phone", NULL,
+			"3", "\\\\phone",					"phone", NULL,						// the correct behavior if '\' is not in charset
+			"3", "pho\\\\ne",					"pho", "ne", NULL,
+			"3", "phon\\\\e",					"phon", NULL,
 			NULL
 		};
 
@@ -151,7 +156,7 @@ void TestTokenizer ( bool bUTF8 )
 		}
 
 		// test misc SBCS-only and UTF8-only one-liners
-		char * dTests2[] =
+		const char * dTests2[] =
 		{
 			"0", "\x80\x81\x82",				"\x80\x81\x82", NULL,
 			"1", "\xC2\x80\xC2\x81\xC2\x82",	"\xC2\x80\xC2\x81\xC2\x82", NULL,
@@ -179,7 +184,7 @@ void TestTokenizer ( bool bUTF8 )
 		if ( bUTF8 )
 		{
 			printf ( "%s for proper UTF-8 error handling\n", sPrefix );
-			char * sLine3 = "hi\xd0\xffh";
+			const char * sLine3 = "hi\xd0\xffh";
 
 			pTokenizer->SetBuffer ( (BYTE*)sLine3, 4 );
 			assert ( !strcmp ( (char*)pTokenizer->GetToken(), "hi" ) );
@@ -189,7 +194,7 @@ void TestTokenizer ( bool bUTF8 )
 		printf ( "%s for uberlong token handling\n", sPrefix );
 
 		const int UBERLONG = 4096;
-		char * sLine4 = new char [ UBERLONG+1 ]; 
+		char * sLine4 = new char [ UBERLONG+1 ];
 		memset ( sLine4, 'a', UBERLONG );
 		sLine4[UBERLONG] = '\0';
 
@@ -204,12 +209,17 @@ void TestTokenizer ( bool bUTF8 )
 		// test short word callbacks
 		printf ( "%s for short token handling\n", sPrefix );
 		ISphTokenizer * pShortTokenizer = pTokenizer->Clone ( bEscaped );
+
 		CSphRemapRange tStar ( '*', '*', '*' );
 		pShortTokenizer->AddCaseFolding ( tStar );
-		pShortTokenizer->SetMinWordLen ( 5 );
+
+		CSphTokenizerSettings tSettings = pShortTokenizer->GetSettings();
+		tSettings.m_iMinWordLen = 5;
+		pShortTokenizer->Setup ( tSettings );
+
 		pShortTokenizer->EnableQueryParserMode ( true );
 
-		char * dTestsShort[] =
+		const char * dTestsShort[] =
 		{
 			"ab*",		"ab*",		NULL,
 			"*ab",		"*ab",		NULL,
@@ -263,7 +273,7 @@ void TestTokenizer ( bool bUTF8 )
 
 		SafeDelete ( sLine4 );
 
-		// test boundaries 
+		// test boundaries
 		printf ( "%s for boundaries handling, run=%d\n", sPrefix, iRun );
 
 		CSphString sError;
@@ -323,18 +333,18 @@ void BenchTokenizer ( bool bUTF8 )
 		const int iPasses = 10;
 		int iTokens = 0;
 
-		float fTime = -sphLongTimer ();
+		int64_t tmTime = -sphMicroTimer();
 		for ( int iPass=0; iPass<iPasses; iPass++ )
 		{
 			pTokenizer->SetBuffer ( (BYTE*)sData, iData );
 			while ( pTokenizer->GetToken() ) iTokens++;
 		}
-		fTime += sphLongTimer ();
+		tmTime += sphMicroTimer();
 
 		iTokens /= iPasses;
-		fTime /= iPasses;
+		tmTime /= iPasses;
 
-		printf ( "run %d: %d bytes, %d tokens, %.3f ms, %.3f MB/sec\n", iRun, iData, iTokens, 1000.0f*fTime, iData/fTime/1000000.0f );
+		printf ( "run %d: %d bytes, %d tokens, %d.%03d ms, %.3f MB/sec\n", iRun, iData, iTokens, int(tmTime/1000), int(tmTime%1000), float(iData)/tmTime );
 		SafeDeleteArray ( sData );
 	}
 }
@@ -356,12 +366,14 @@ void TestStripper ()
 		{ "testing <1 <\" <\x80 <\xe0 <\xff </3 malformed tags", "", "", "testing <1 <\" <\x80 <\xe0 <\xff </3 malformed tags" },
 		{ "testing comm<!--comm-->ents", "", "", "testing comments" },
 		{ "&lt; &gt; &thetasym; &somethingverylong; &the", "", "", "< > \xCF\x91 &somethingverylong; &the" },
-		{ "testing <img src=\"g/smth.jpg\" alt=\"nice picture\" rel=anotherattr junk=throwaway>inline tags vs attr indexing", "img=alt,rel", "", "testing nice picture anotherattr inline tags vs attr indexing" }
+		{ "testing <img src=\"g/smth.jpg\" alt=\"nice picture\" rel=anotherattr junk=throwaway>inline tags vs attr indexing", "img=alt,rel", "", "testing nice picture anotherattr inline tags vs attr indexing" },
+		{ "this <?php $code = \"must be stripped\"; ?> away", "", "", "this  away" }
 	};
 
-	for ( int iTest=0; iTest<(int)(sizeof(sTests)/sizeof(sTests[0])); iTest++ )
+	int nTests = (int)(sizeof(sTests)/sizeof(sTests[0]));
+	for ( int iTest=0; iTest<nTests; iTest++ )
 	{
-		printf ( "testing HTML stripper, test %d\n", 1+iTest );
+		printf ( "testing HTML stripper, test %d/%d... ", 1+iTest, nTests );
 
 		CSphString sError;
 		CSphHTMLStripper tStripper;
@@ -371,6 +383,8 @@ void TestStripper ()
 		CSphString sBuf ( sTests[iTest][0] );
 		tStripper.Strip ( (BYTE*)sBuf.cstr() );
 		assert ( strcmp ( sBuf.cstr(), sTests[iTest][3] )==0 );
+
+		printf ( "ok\n" );
 	}
 }
 
@@ -402,15 +416,16 @@ void BenchStripper ()
 			tStripper.SetRemovedElements ( "style, script", sError );
 
 		const int iPasses = 50;
-		float fTime = -sphLongTimer ();
+		int64_t tmTime = -sphMicroTimer();
 		for ( int iPass=0; iPass<iPasses; iPass++ )
 		{
 			tStripper.Strip ( (BYTE*)sBuf );
 			strcpy ( sBuf, sRef );
 		}
-		fTime += sphLongTimer ();
+		tmTime += sphMicroTimer();
 
-		printf ( "run %d: %d bytes, %.3f ms, %.3f MB/sec\n", iRun, iLen, 1000.0f*fTime/float(iPasses), iLen*float(iPasses)/fTime/1000000.0f );
+		tmTime /= iPasses;
+		printf ( "run %d: %d bytes, %d.%03d ms, %.3f MB/sec\n", iRun, iLen, int(tmTime/1000), int(tmTime%1000), float(iLen)/tmTime );
 	}
 
 	SafeDeleteArray ( sBuf );
@@ -438,7 +453,7 @@ void TestExpr ()
 		tMatch.m_pRowitems[i] = 1+i;
 
 	struct ExprTest_t
-	{ 
+	{
 		const char *	m_sExpr;
 		float			m_fValue;
 	};
@@ -472,14 +487,13 @@ void TestExpr ()
 		{ "-aaa>-bbb",						1.0f },
 	};
 
-	const int iTests = sizeof(dTests)/sizeof(dTests[0]);
-	for ( int iTest=0; iTest<iTests; iTest++ )
+	const int nTests = sizeof(dTests)/sizeof(dTests[0]);
+	for ( int iTest=0; iTest<nTests; iTest++ )
 	{
-		printf ( "testing expression evaluation, test %d/%d... ", 1+iTest, iTests );
+		printf ( "testing expression evaluation, test %d/%d... ", 1+iTest, nTests );
 
 		CSphString sError;
-		bool bCalcGeoDist;
-		CSphScopedPtr<ISphExpr> pExpr ( sphExprParse ( dTests[iTest].m_sExpr, tSchema, bCalcGeoDist, sError ) );
+		CSphScopedPtr<ISphExpr> pExpr ( sphExprParse ( dTests[iTest].m_sExpr, tSchema, NULL, NULL, sError ) );
 		if ( !pExpr.Ptr() )
 		{
 			printf ( "FAILED; %s\n", sError.cstr() );
@@ -508,7 +522,7 @@ void TestExpr ()
 #define BBB float(tMatch.m_pRowitems[1])
 #define CCC float(tMatch.m_pRowitems[2])
 
-NOINLINE float ExprNative1 ( const CSphMatch & tMatch )	{ return AAA+BBB*CCC-0.75f;}
+NOINLINE float ExprNative1 ( const CSphMatch & tMatch )	{ return AAA+BBB*CCC-1.0f;}
 NOINLINE float ExprNative2 ( const CSphMatch & tMatch )	{ return AAA+BBB*CCC*2.0f-3.0f/4.0f*5.0f/6.0f*BBB; }
 NOINLINE float ExprNative3 ( const CSphMatch & )		{ return (float)sqrt ( 2.0f ); }
 
@@ -540,7 +554,7 @@ void BenchExpr ()
 	};
 	ExprBench_t dBench[] =
 	{
-		{ "aaa+bbb*(ccc)-0.75",				ExprNative1 },
+		{ "aaa+bbb*(ccc)-1",				ExprNative1 },
 		{ "aaa+bbb*ccc*2-3/4*5/6*bbb",		ExprNative2 },
 		{ "sqrt(2)",						ExprNative3 }
 	};
@@ -549,9 +563,9 @@ void BenchExpr ()
 	{
 		printf ( "run %d: ", iRun+1 );
 
+		DWORD uType;
 		CSphString sError;
-		bool bCalcGeoDist;
-		CSphScopedPtr<ISphExpr> pExpr ( sphExprParse ( dBench[iRun].m_sExpr, tSchema, bCalcGeoDist, sError ) );
+		CSphScopedPtr<ISphExpr> pExpr ( sphExprParse ( dBench[iRun].m_sExpr, tSchema, &uType, NULL, sError ) );
 		if ( !pExpr.Ptr() )
 		{
 			printf ( "FAILED; %s\n", sError.cstr() );
@@ -561,45 +575,68 @@ void BenchExpr ()
 		const int NRUNS = 1000000;
 
 		volatile float fValue = 0.0f;
-		float fTime = sphLongTimer ();
+		int64_t tmTime = sphMicroTimer();
 		for ( int i=0; i<NRUNS; i++ ) fValue += pExpr->Eval(tMatch);
-		fTime = sphLongTimer() - fTime;
+		tmTime = sphMicroTimer() - tmTime;
 
-		float fTimeNative = sphLongTimer ();
+		int64_t tmTimeInt = sphMicroTimer();
+		if ( uType==SPH_ATTR_INTEGER )
+		{
+			int uValue = 0;
+			for ( int i=0; i<NRUNS; i++ ) uValue += pExpr->IntEval(tMatch);
+		}
+		tmTimeInt = sphMicroTimer() - tmTimeInt;
+
+		int64_t tmTimeNative = sphMicroTimer();
 		for ( int i=0; i<NRUNS; i++ ) fValue += dBench[iRun].m_pFunc ( tMatch );
-		fTimeNative = sphLongTimer() - fTimeNative;
+		tmTimeNative = sphMicroTimer() - tmTimeNative;
 
-		printf ( "interpreted %.1f Mcalls/sec, native %.1f Mcalls/sec\n",
-			float(NRUNS)/float(1000000.0f)/fTime,
-			float(NRUNS)/float(1000000.0f)/fTimeNative );
+		if ( uType==SPH_ATTR_INTEGER )
+			printf ( "int-eval %.1fM/sec, ", float(NRUNS)/tmTimeInt );
+
+		printf ( "flt-eval %.1fM/sec, native %.1fM/sec\n",
+			float(NRUNS)/tmTime,
+			float(NRUNS)/tmTimeNative );
 	}
 }
 
 //////////////////////////////////////////////////////////////////////////
 
-CSphString ReconstructNode ( const CSphExtendedQueryNode * pNode )
+CSphString ReconstructNode ( const XQNode_t * pNode, const CSphSchema & tSchema )
 {
 	CSphString sRes ( "" );
 
 	if ( pNode->IsPlain() )
 	{
 		// say just words to me
-		const CSphExtendedQueryAtom & tAtom = pNode->m_tAtom;
-		const CSphVector<CSphExtendedQueryAtomWord> & dWords = tAtom.m_dWords;
+		const CSphVector<XQKeyword_t> & dWords = pNode->m_dWords;
 		ARRAY_FOREACH ( i, dWords )
 			sRes.SetSprintf ( "%s %s", sRes.cstr(), dWords[i].m_sWord.cstr() );
 		sRes.Chop ();
 
-		if ( tAtom.m_bQuorum || tAtom.m_iMaxDistance>0 )
+		if ( pNode->m_bQuorum || pNode->m_iMaxDistance>0 )
 		{
-			sRes.SetSprintf ( "\"%s\"%c%d", sRes.cstr(), tAtom.m_bQuorum ? '/' : '~', tAtom.m_iMaxDistance ); // quorum or proximity
-		
+			sRes.SetSprintf ( "\"%s\"%c%d", sRes.cstr(), pNode->m_bQuorum ? '/' : '~', pNode->m_iMaxDistance ); // quorum or proximity
 		} else if ( dWords.GetLength()>1 )
 		{
-			if ( tAtom.m_iMaxDistance==0 )
+			if ( pNode->m_iMaxDistance==0 )
 				sRes.SetSprintf ( "\"%s\"", sRes.cstr() ); // phrase
 			else
-				sRes.SetSprintf ( "( %s )", sRes.cstr() ); // just bag of words
+				sRes.SetSprintf ( "%s", sRes.cstr() ); // just bag of words
+		}
+
+		if ( pNode->m_uFieldMask!=0xFFFFFFFFUL )
+		{
+			CSphString sFields ( "" );
+			for ( int i=0; i<32; i++ )
+				if ( pNode->m_uFieldMask & (1<<i) )
+					sFields.SetSprintf ( "%s,%s", sFields.cstr(), tSchema.m_dFields[i].m_sName.cstr() );
+
+			sRes.SetSprintf ( "( @%s: %s )", sFields.cstr()+1, sRes.cstr() );
+		} else
+		{
+			if ( !pNode->m_bQuorum && pNode->m_iMaxDistance<0 && dWords.GetLength()>1 )
+				sRes.SetSprintf ( "( %s )", sRes.cstr() ); // wrap bag of words
 		}
 
 	} else
@@ -607,9 +644,20 @@ CSphString ReconstructNode ( const CSphExtendedQueryNode * pNode )
 		ARRAY_FOREACH ( i, pNode->m_dChildren )
 		{
 			if ( !i )
-				sRes = ReconstructNode ( pNode->m_dChildren[i] );
+				sRes = ReconstructNode ( pNode->m_dChildren[i], tSchema );
 			else
-				sRes.SetSprintf ( "%s %s %s", sRes.cstr(), pNode->m_bAny ? "OR" : "AND", ReconstructNode ( pNode->m_dChildren[i] ).cstr() );
+			{
+				const char * sOp = "(unknown-op)";
+				switch ( pNode->m_eOp )
+				{
+					case SPH_QUERY_AND:		sOp = "AND"; break;
+					case SPH_QUERY_OR:		sOp = "OR"; break;
+					case SPH_QUERY_NOT:		sOp = "NOT"; break;
+					case SPH_QUERY_ANDNOT:	sOp = "AND NOT"; break;
+					case SPH_QUERY_BEFORE:	sOp = "BEFORE"; break;
+				}
+				sRes.SetSprintf ( "%s %s %s", sRes.cstr(), sOp, ReconstructNode ( pNode->m_dChildren[i], tSchema ).cstr() );
+			}
 		}
 
 		if ( pNode->m_dChildren.GetLength()>1 )
@@ -619,16 +667,6 @@ CSphString ReconstructNode ( const CSphExtendedQueryNode * pNode )
 	return sRes;
 }
 
-CSphString ReconstructQuery ( const CSphExtendedQuery & tQuery )
-{
-	CSphString sAccept = ReconstructNode ( tQuery.m_pAccept );
-	CSphString sReject = ReconstructNode ( tQuery.m_pReject );
-
-	if ( !sReject.IsEmpty () )
-		sAccept.SetSprintf ( "( %s ) AND NOT ( %s )", sAccept.cstr(), sReject.cstr() );
-
-	return sAccept;
-}
 
 void TestQueryParser ()
 {
@@ -637,13 +675,23 @@ void TestQueryParser ()
 	CSphSchema tSchema;
 	CSphColumnInfo tCol;
 	tCol.m_sName = "title"; tSchema.m_dFields.Add ( tCol );
-	tCol.m_sName = "content"; tSchema.m_dFields.Add ( tCol );
+	tCol.m_sName = "body"; tSchema.m_dFields.Add ( tCol );
 
+	CSphDictSettings tDictSettings;
 	CSphScopedPtr<ISphTokenizer> pTokenizer ( sphCreateSBCSTokenizer () );
-	CSphScopedPtr<CSphDict> pDict ( sphCreateDictionaryCRC ( NULL, NULL, NULL, pTokenizer.Ptr(), sTmp ) );
+	CSphScopedPtr<CSphDict> pDict ( sphCreateDictionaryCRC ( tDictSettings, pTokenizer.Ptr(), sTmp ) );
 	assert ( pTokenizer.Ptr() );
 	assert ( pDict.Ptr() );
 
+	CSphTokenizerSettings tTokenizerSetup;
+	tTokenizerSetup.m_iMinWordLen = 2;
+	tTokenizerSetup.m_sSynonymsFile = g_sTmpfile;
+	pTokenizer->Setup ( tTokenizerSetup );
+
+	CSphString sError;
+	assert ( CreateSynonymsFile ( NULL ) );
+	assert ( pTokenizer->LoadSynonyms ( g_sTmpfile, sError ) );
+
 	struct QueryTest_t
 	{
 		const char *	m_sQuery;
@@ -656,23 +704,44 @@ void TestQueryParser ()
 		{ "aaa bbb|ccc",					"( aaa AND ( bbb OR ccc ) )" },
 		{ "aaa (bbb ccc)|ddd",				"( aaa AND ( ( bbb AND ccc ) OR ddd ) )" },
 		{ "aaa bbb|(ccc ddd)",				"( aaa AND ( bbb OR ( ccc AND ddd ) ) )" },
-		{ "aaa bbb|(ccc ddd)|eee|(fff)",	"( aaa AND ( ( ( bbb OR ( ccc AND ddd ) ) OR eee ) OR fff ) )" },
-		{ "aaa bbb|(ccc ddd) eee|(fff)",	"( ( aaa AND ( bbb OR ( ccc AND ddd ) ) ) AND ( eee OR fff ) )" },
-		{ "aaa (ccc ddd)|bbb|eee|(fff)",	"( aaa AND ( ( ( ( ccc AND ddd ) OR bbb ) OR eee ) OR fff ) )" },
-		{ "aaa (ccc ddd)|bbb eee|(fff)",	"( ( aaa AND ( ( ccc AND ddd ) OR bbb ) ) AND ( eee OR fff ) )" },
+		{ "aaa bbb|(ccc ddd)|eee|(fff)",	"( aaa AND ( bbb OR ( ccc AND ddd ) OR eee OR fff ) )" },
+		{ "aaa bbb|(ccc ddd) eee|(fff)",	"( aaa AND ( bbb OR ( ccc AND ddd ) ) AND ( eee OR fff ) )" },
+		{ "aaa (ccc ddd)|bbb|eee|(fff)",	"( aaa AND ( ( ccc AND ddd ) OR bbb OR eee OR fff ) )" },
+		{ "aaa (ccc ddd)|bbb eee|(fff)",	"( aaa AND ( ( ccc AND ddd ) OR bbb ) AND ( eee OR fff ) )" },
 		{ "aaa \"bbb ccc\"~5|ddd",			"( aaa AND ( \"bbb ccc\"~5 OR ddd ) )" },
 		{ "aaa bbb|\"ccc ddd\"~5",			"( aaa AND ( bbb OR \"ccc ddd\"~5 ) )" },
+		{ "aaa ( ( \"bbb ccc\"~3|ddd ) eee | ( fff -ggg ) )",	"( aaa AND ( ( \"bbb ccc\"~3 OR ddd ) AND ( eee OR ( fff AND NOT ggg ) ) ) )" },
+		{ "@title aaa @body ccc|(@title ddd eee)|fff ggg",		"( ( @title: aaa ) AND ( ( @body: ccc ) OR ( ( @title: ddd ) AND ( @title: eee ) ) OR ( @body: fff ) ) AND ( @body: ggg ) )" },
+		{ "@title hello world | @body sample program",			"( ( @title: hello ) AND ( ( @title: world ) OR ( @body: sample ) ) AND ( @body: program ) )" },
+		{ "@title one two three four",							"( ( @title: one ) AND ( @title: two ) AND ( @title: three ) AND ( @title: four ) )" },
+		{ "@title one (@body two three) four",					"( ( @title: one ) AND ( ( @body: two ) AND ( @body: three ) ) AND ( @title: four ) )" },
+		{ "windows 7 2000",										"( windows AND 2000 )" },
+		{ "aaa a|bbb",											"( aaa AND bbb )" },
+		{ "aaa bbb|x y z|ccc",									"( aaa AND bbb AND ccc )" },
+		{ "a",													"" },
+		{ "hello -world",										"( hello AND NOT world )" },
+		{ "-hello world",										"( world AND NOT hello )" },
+		{ "\"phrase (query)/3 ~on steroids\"",					"\"phrase query on steroids\"" },
+		{ "hello a world",										"( hello AND world )" },
+		{ "-one",												"" },
+		{ "-one -two",											"" },
+		{ "\"\"",												"" },
+		{ "\"()\"",												"" },
+		{ "\"]\"",												"" },
+		{ "@title hello @body -world",							"( ( @title: hello ) AND NOT ( @body: world ) )" },
+		{ "Ms-Dos",							"MS-DOS" }
 	};
 
+
 	int nTests = sizeof(dTest)/sizeof(dTest[0]);
 	for ( int i=0; i<nTests; i++ )
 	{
 		printf ( "testing query parser, test %d/%d... ", i+1, nTests );
 
-		CSphExtendedQuery tQuery;
+		XQQuery_t tQuery;
 		sphParseExtendedQuery ( tQuery, dTest[i].m_sQuery, pTokenizer.Ptr(), &tSchema, pDict.Ptr() );
 
-		CSphString sReconst = ReconstructQuery ( tQuery );
+		CSphString sReconst = ReconstructNode ( tQuery.m_pRoot, tSchema );
 		assert ( sReconst==dTest[i].m_sReconst );
 
 		printf ( "ok\n" );
@@ -704,5 +773,5 @@ int main ()
 }
 
 //
-// $Id: tests.cpp 1434 2008-09-09 15:35:07Z shodan $
+// $Id: tests.cpp 1716 2009-03-02 15:42:35Z klirichek $
 //
